@@ -2,7 +2,7 @@
 
 import os, sys, logging, subprocess
 import gi
-from cameractrls import CameraCtrls, find_by_text_id, get_devices, v4ldirs, find_idx
+from cameractrls import CameraCtrls, PTZHWControllers, find_by_text_id, get_devices, v4ldirs, find_idx
 from cameractrls import version, ghurl
 
 gi.require_version('Gtk', '4.0')
@@ -19,14 +19,17 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
         self.device = None
         self.camera = None
         self.listener = None
+        self.ptz_controllers = None
 
         self.grid = None
         self.frame = None
         self.device_dd = None
 
-        self.zoom_absolute = None
+        self.zoom_absolute_sc = None
         self.pan_speed_sc = None
         self.tilt_speed_sc = None
+        self.pan_absolute_sc = None
+        self.tilt_absolute_sc = None
 
         self.init_window()
         self.refresh_devices()
@@ -73,27 +76,39 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
             Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
         )
 
-        hambuger_menu = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=10)
-        hambuger_menu.append(Gtk.Button(action_name='app.about', label='About', has_frame=False))
-
-        hamburger_button = Gtk.MenuButton(
-            popover=Gtk.Popover(position=Gtk.PositionType.BOTTOM, child=hambuger_menu),
-            icon_name='open-menu-symbolic', has_frame=False,
+        about_button = Gtk.Button(
+            action_name='app.about',
+            icon_name='info-symbolic',
+            has_frame=False,
+            tooltip_text='About',
         )
 
         self.open_cam_button = Gtk.Button(
             action_name='app.open_camera_window',
             action_target=GLib.Variant('s', ''),
             icon_name='camera-video-symbolic', has_frame=False,
+            tooltip_text='Show preview window',
         )
 
-        refresh_button = Gtk.Button(icon_name='view-refresh-symbolic', has_frame=False)
+        self.ptz_model = Gtk.StringList()
+        self.ptz_lb = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
+        self.ptz_lb.bind_model(self.ptz_model, lambda name: Gtk.ToggleButton(label=name.get_string()))
+
+        self.ptz_sw = Gtk.MenuButton(
+            popover=Gtk.Popover(position=Gtk.PositionType.BOTTOM, child=self.ptz_lb),
+            icon_name='input-gaming-symbolic',
+            has_frame=False,
+            tooltip_text='PTZ hardware controls',
+        )
+
+        refresh_button = Gtk.Button(icon_name='view-refresh-symbolic', has_frame=False, tooltip_text='Refresh')
         refresh_button.connect('clicked', lambda e: self.refresh_devices())
 
         headerbar = Gtk.HeaderBar(show_title_buttons=True)
         headerbar.pack_start(refresh_button)
-        headerbar.pack_end(hamburger_button)
+        headerbar.pack_end(about_button)
         headerbar.pack_end(self.open_cam_button)
+        headerbar.pack_end(self.ptz_sw)
         self.set_titlebar(headerbar)
 
         self.grid = Gtk.Grid()
@@ -107,7 +122,10 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
             self.zero_box.append(Gtk.Label(label='Please permit access with', max_width_chars=30, wrap=True, margin_top=10))
             self.zero_box.append(Gtk.Label(label='snap connect cameractrls:camera', selectable=True, margin_bottom=10))
 
-        self.device_dd = Gtk.DropDown(model=Gtk.StringList(), hexpand=True, halign=Gtk.Align.START, show_arrow=False)
+        self.device_dd = Gtk.DropDown(
+            model=Gtk.StringList(), hexpand=True, halign=Gtk.Align.START, show_arrow=False,
+            tooltip_text='Select other camera',
+        )
         # use the default factory as list_factory
         self.device_dd.set_list_factory(self.device_dd.get_factory())
         self.device_dd.get_first_child().set_has_frame(False)
@@ -194,6 +212,19 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
             lambda errs: GLib.idle_add(self.notify, '\n'.join(errs)),
         )
         self.device = device
+        self.ptz_controllers = PTZHWControllers(self.device.path,
+            lambda check, p, i: GLib.timeout_add(300, check, p, i),
+            lambda err: self.notify(err),
+            lambda i: self.ptz_lb.get_row_at_index(i).get_child().set_active(False),
+        )
+        self.ptz_model.splice(0, self.ptz_model.get_n_items(), self.ptz_controllers.get_names())
+        for i in range(self.ptz_model.get_n_items()):
+            row = self.ptz_lb.get_row_at_index(i)
+            row.set_activatable(False)
+            row.set_focusable(False)
+            row.get_child().connect('toggled', lambda c, i=i: self.ptz_controllers.set_active(i, c.get_active()))
+        self.ptz_sw.set_visible(self.camera.has_ptz())
+        self.ptz_sw.set_sensitive(self.ptz_model.get_n_items() != 0)
         self.open_cam_button.set_action_target_value(GLib.Variant('s', self.device.path))
 
     def close_device(self):
@@ -203,6 +234,8 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
             self.camera = None
             self.listener.stop()
             self.listener = None
+            self.ptz_controllers.terminate_all()
+            self.ptz_controllers = None
             os.close(self.fd)
             self.fd = 0
 
@@ -282,11 +315,9 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
                             for controller in scale.observe_controllers():
                                 if isinstance(controller, gi.repository.Gtk.GestureClick):
                                     controller.connect('released', lambda c, n, x, y, sc=scale: sc.set_value(0))
-                                if isinstance(controller, gi.repository.Gtk.ShortcutController):
-                                    scale.remove_controller(controller)
                                 if isinstance(controller, gi.repository.Gtk.EventControllerKey):
-                                    controller.connect('key-pressed', self.handle_ptz_key_pressed)
-                                    controller.connect('key-released', self.handle_ptz_key_released)
+                                    controller.connect('key-pressed', self.handle_ptz_speed_key_pressed)
+                                    controller.connect('key-released', self.handle_ptz_speed_key_released)
                         if c.step and c.step != 1:
                             adjustment_step = Gtk.Adjustment(lower=c.min, upper=c.max, value=c.value, step_increment=c.step)
                             adjustment_step.connect('value-changed', lambda a,c=c,a1=adjustment: [a.set_value(a.get_value() - a.get_value() % c.step),a1.set_value(a.get_value())])
@@ -310,6 +341,17 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
                         if c.text_id == 'tilt_speed':
                             self.tilt_speed_sc = scale
 
+                        if c.text_id == 'pan_absolute':
+                            self.pan_absolute_sc = scale
+
+                        if c.text_id == 'tilt_absolute':
+                            self.tilt_absolute_sc = scale
+
+                        if c.text_id in ['pan_absolute', 'tilt_absolute']:
+                            for controller in scale.observe_controllers():
+                                if isinstance(controller, gi.repository.Gtk.EventControllerKey):
+                                    controller.connect('key-pressed', self.handle_ptz_absolute_key_pressed)
+
                         refresh = Gtk.Button(icon_name='edit-undo-symbolic', valign=Gtk.Align.CENTER, halign=Gtk.Align.START, has_frame=False)
                         refresh.connect('clicked', lambda e, c=c, sc=scale: sc.get_adjustment().set_value(c.default))
                         ctrl_box.append(refresh)
@@ -353,6 +395,9 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
                                     self.update_ctrl(c, m.lp_text_id)
                                 ])
                                 b.add_controller(lp)
+                            if m.use_shortcut:
+                                b.set_label('_' + m.name)
+                                b.set_use_underline(True)
                             box.append(b)
                             c.gui_ctrls += b
                         if c.default is not None:
@@ -466,7 +511,7 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
             c.gui_value_set(c.value)
         self.update_ctrl_state(c)
 
-    def handle_ptz_key_pressed(self, c, keyval, keycode, state):
+    def handle_ptz_speed_key_pressed(self, c, keyval, keycode, state):
         pan_lower = self.pan_speed_sc.get_adjustment().get_lower()
         pan_upper =self.pan_speed_sc.get_adjustment().get_upper()
         tilt_lower = self.tilt_speed_sc.get_adjustment().get_lower()
@@ -493,31 +538,74 @@ class CameraCtrlsWindow(Gtk.ApplicationWindow):
             self.pan_speed_sc.set_value(pan_upper)
             self.tilt_speed_sc.set_value(tilt_upper)
         elif self.zoom_absolute_sc is not None:
-            zoom_value = self.zoom_absolute_sc.get_value()
-            zoom_step = self.zoom_absolute_sc.get_adjustment().get_step_increment()
-            zoom_page = self.zoom_absolute_sc.get_adjustment().get_page_increment()
-            zoom_lower = self.zoom_absolute_sc.get_adjustment().get_lower()
-            zoom_upper = self.zoom_absolute_sc.get_adjustment().get_upper()
-
-            if keyval in [Gdk.KEY_plus, Gdk.KEY_KP_Add] and state == Gdk.ModifierType.CONTROL_MASK:
-                self.zoom_absolute_sc.set_value(zoom_value + zoom_page)
-            elif keyval in [Gdk.KEY_minus, Gdk.KEY_KP_Subtract] and state == Gdk.ModifierType.CONTROL_MASK:
-                self.zoom_absolute_sc.set_value(zoom_value - zoom_page)
-            elif keyval in [Gdk.KEY_plus, Gdk.KEY_KP_Add]:
-                self.zoom_absolute_sc.set_value(zoom_value + zoom_step)
-            elif keyval in [Gdk.KEY_minus, Gdk.KEY_KP_Subtract]:
-                self.zoom_absolute_sc.set_value(zoom_value - zoom_step)
-            elif keyval in [Gdk.KEY_Home]:
-                self.zoom_absolute_sc.set_value(zoom_lower)
-            elif keyval in [Gdk.KEY_End]:
-                self.zoom_absolute_sc.set_value(zoom_upper)
-            else:
-                return False
+            return self.handle_ptz_key_pressed_zoom(keyval, state)
         else:
             return False
         return True
 
-    def handle_ptz_key_released(self, c, keyval, keycode, state):
+    def handle_ptz_absolute_key_pressed(self, c, keyval, keycode, state):
+        pan_value = self.pan_absolute_sc.get_value()
+        pan_step = self.pan_absolute_sc.get_adjustment().get_step_increment()
+        tilt_value = self.tilt_absolute_sc.get_value()
+        tilt_step = self.tilt_absolute_sc.get_adjustment().get_step_increment()
+    
+        if keyval in [Gdk.KEY_Left, Gdk.KEY_KP_Left, Gdk.KEY_a]:
+            self.pan_absolute_sc.set_value(pan_value - pan_step)
+        elif keyval in [Gdk.KEY_Right, Gdk.KEY_KP_Right, Gdk.KEY_d]:
+            self.pan_absolute_sc.set_value(pan_value + pan_step)
+        elif keyval in [Gdk.KEY_Up, Gdk.KEY_KP_Up, Gdk.KEY_w]:
+            self.tilt_absolute_sc.set_value(tilt_value + tilt_step)
+        elif keyval in [Gdk.KEY_Down, Gdk.KEY_KP_Down, Gdk.KEY_s]:
+            self.tilt_absolute_sc.set_value(tilt_value - tilt_step)
+        elif keyval == Gdk.KEY_KP_End:
+            self.pan_absolute_sc.set_value(pan_value - pan_step)
+            self.tilt_absolute_sc.set_value(tilt_value - tilt_step)
+        elif keyval == Gdk.KEY_KP_Page_Down:
+            self.pan_absolute_sc.set_value(pan_value + pan_step)
+            self.tilt_absolute_sc.set_value(tilt_value - tilt_step)
+        elif keyval == Gdk.KEY_KP_Home:
+            self.pan_absolute_sc.set_value(pan_value - pan_step)
+            self.tilt_absolute_sc.set_value(tilt_value + tilt_step)
+        elif keyval == Gdk.KEY_KP_Page_Up:
+            self.pan_absolute_sc.set_value(pan_value + pan_step)
+            self.tilt_absolute_sc.set_value(tilt_value + tilt_step)
+        elif keyval == Gdk.KEY_KP_Insert:
+            self.pan_absolute_sc.set_value(0)
+            self.tilt_absolute_sc.set_value(0)
+        elif self.zoom_absolute_sc is not None:
+            return self.handle_ptz_key_pressed_zoom(keyval, state)
+        else:
+            return False
+        return True
+
+    def handle_ptz_key_pressed_zoom(self, keyval, state):
+        zoom_value = self.zoom_absolute_sc.get_value()
+        zoom_step = self.zoom_absolute_sc.get_adjustment().get_step_increment()
+        zoom_page = self.zoom_absolute_sc.get_adjustment().get_page_increment()
+        zoom_lower = self.zoom_absolute_sc.get_adjustment().get_lower()
+        zoom_upper = self.zoom_absolute_sc.get_adjustment().get_upper()
+
+        if keyval in [Gdk.KEY_plus, Gdk.KEY_KP_Add] and state & Gdk.ModifierType.CONTROL_MASK:
+            self.zoom_absolute_sc.set_value(zoom_value + zoom_page)
+        elif keyval in [Gdk.KEY_minus, Gdk.KEY_KP_Subtract] and state & Gdk.ModifierType.CONTROL_MASK:
+            self.zoom_absolute_sc.set_value(zoom_value - zoom_page)
+        elif keyval in [Gdk.KEY_plus, Gdk.KEY_KP_Add]:
+            self.zoom_absolute_sc.set_value(zoom_value + zoom_step)
+        elif keyval in [Gdk.KEY_minus, Gdk.KEY_KP_Subtract]:
+            self.zoom_absolute_sc.set_value(zoom_value - zoom_step)
+        elif keyval in [Gdk.KEY_Home]:
+            self.zoom_absolute_sc.set_value(zoom_lower)
+        elif keyval in [Gdk.KEY_End]:
+            self.zoom_absolute_sc.set_value(zoom_upper)
+        elif keyval in [Gdk.KEY_Page_Up]:
+            self.zoom_absolute_sc.set_value(zoom_value + zoom_page)
+        elif keyval in [Gdk.KEY_Page_Down]:
+            self.zoom_absolute_sc.set_value(zoom_value - zoom_page)
+        else:
+            return False
+        return True
+
+    def handle_ptz_speed_key_released(self, c, keyval, keycode, state):
         if keyval in [Gdk.KEY_Left, Gdk.KEY_Right, Gdk.KEY_KP_Left, Gdk.KEY_KP_Right, Gdk.KEY_a, Gdk.KEY_d]:
             self.pan_speed_sc.set_value(0)
         elif keyval in [Gdk.KEY_Up, Gdk.KEY_Down, Gdk.KEY_KP_Up, Gdk.KEY_KP_Down, Gdk.KEY_w, Gdk.KEY_s]:
@@ -565,7 +653,7 @@ class CameraCtrlsApp(Gtk.Application):
         about.set_program_name('Cameractrls')
         about.set_authors(['Gergo Koteles <soyer@irl.hu>'])
         about.set_artists(['Jorge Toledo https://lacajita.es'])
-        about.set_copyright('Copyright © 2022 - 2023 Gergo Koteles')
+        about.set_copyright('Copyright © 2022 - 2024 Gergo Koteles')
         about.set_license_type(Gtk.License.MIT_X11)
         about.set_logo_icon_name('hu.irl.cameractrls')
         about.set_website(ghurl)
